import { NormalizedScalarsMap, ParsedEnumValuesMap } from "@graphql-codegen/visitor-plugin-common";
import { GraphQLSchema, SelectionSetNode, GraphQLNamedType, GraphQLObjectType, FieldNode, GraphQLField, SelectionNode, FragmentSpreadNode, FragmentDefinitionNode, GraphQLType } from "graphql";
import { PLVariableInfo } from "./types";
/** Resolve AST FieldNodes to ResolvedFields. Throws an error if the
 * field is not a field on the parent type.
 */
export declare function resolveFieldNodes(parent: GraphQLObjectType, fields: ReadonlyArray<FieldNode>): ResolvedField[];
/** Resolve a list of SelectionNodes (defined in operations) against
 * a parent and the overall schema.
 */
export declare function resolveSelectionSet(context: ResolveContext, parentType: GraphQLObjectType, set: ReadonlyArray<SelectionNode>): ResolvedSelectionSet;
/** Use fragment name on FragmentSpreadNode to create a list of resolved fields.
 * FragmentSpreadNodes are found in operation definitions and represent the use
 * of a spread vs the definition of a spread.
 *
 * @see resolveFragmentNameToFields
 */
export declare function resolveFragmentSpreadToFields(context: ResolveContext, parentType: GraphQLObjectType, spread: FragmentSpreadNode): ReadonlyArray<ResolvedField>;
/** Give a FieldNode, defined in selection set, return its alias or its name. */
export declare function aliasOrName(f: FieldNode): string;
/** Given the name of a fragment and the overall schema, expand
 * the fragment to a set of fields via the FragmentDefinition found in the context.
 * If any of the toplevel fragment
 * elements are fragments, expand those. Also return the corresponding FieldNode
 * which have additional information about aliases and directives.
 */
export declare function resolveFragmentNameToFields(context: ResolveContext, name: string): [GraphQLNamedType, ReadonlyArray<[FieldNode, GraphQLField<any, any, {}>]>];
/** Given a resolved field, resolve its selections. */
export declare function resolveFieldSelectionSet(context: ResolveContext, field: ResolvedField): ResolvedSelectionSet;
/** Resolve selection sets to a hierarchical, intermediate data structure
 * useful for different target languages and generate scala code. This
 * function is typically used at the highest level of definition
 * for an operation and the parentSchemaType is typically Query, Mutation
 * or Subscription. It is a wrapper over `resolveSelectionSet`.
 */
export declare function selectionSetToObject(context: ResolveContext, parentSchemaType: GraphQLNamedType, selectionSet: SelectionSetNode): ResolvedSelectionSet;
/** For all resolution, basic data that may be needed. */
export declare class ResolveContext {
    schema: GraphQLSchema;
    fragments: ReadonlyArray<FragmentDefinitionNode>;
    scalars: NormalizedScalarsMap;
    constructor(schema: GraphQLSchema, fragments: ReadonlyArray<FragmentDefinitionNode>, scalars: NormalizedScalarsMap);
    /** Find a fragment by name. */
    fragmentByName: (name: string) => FragmentDefinitionNode;
}
/** Intermediate representation of resolved fields for a given
 * type. Sub-objects can be resolved by resolving the "complex"
 * fields.
 */
export declare class ResolvedSelectionSet {
    schema: GraphQLSchema;
    ofType: GraphQLObjectType;
    fields: ReadonlyArray<ResolvedField>;
    constructor(schema: GraphQLSchema, ofType: GraphQLObjectType, fields: ReadonlyArray<ResolvedField>);
    get leaves(): ResolvedField[];
    get complex(): ResolvedField[];
    /** Partition the resolved fields into leaves and non-leaves. */
    get partition(): ResolvedField[][];
    /** This should resolve all objects and their resolved selection sets. */
    resolveComplex: (context: ResolveContext) => [ResolvedField, ResolvedSelectionSet][];
}
/** Intermediate represetation of a field selection from an object for use
 * in an operation such as a query.
 */
export declare class ResolvedField {
    /** Name of field may be influenced by alias and not what "field" says. */
    name: string;
    /** Field type. */
    field: GraphQLField<any, any, {}>;
    /** Parent type. */
    ofType: GraphQLObjectType;
    /** Subfields, if any in AST format. They are not resolved. */
    selectionSet: ReadonlyArray<SelectionNode>;
    constructor(
    /** Name of field may be influenced by alias and not what "field" says. */
    name: string, 
    /** Field type. */
    field: GraphQLField<any, any, {}>, 
    /** Parent type. */
    ofType: GraphQLObjectType, 
    /** Subfields, if any in AST format. They are not resolved. */
    selectionSet: ReadonlyArray<SelectionNode>);
    /** Same as constructor. */
    static from: (ofType: GraphQLObjectType<any, any, {
        [key: string]: any;
    }>, field: GraphQLField<any, any, {}>, name: string, selectionSet: readonly SelectionNode[]) => ResolvedField;
    /** FieldNodes are defined in operations. */
    static fromField: (ofType: GraphQLObjectType<any, any, {
        [key: string]: any;
    }>, field: GraphQLField<any, any, {}>, astNode: FieldNode) => ResolvedField;
    get isLeaf(): boolean;
    get isScalar(): boolean;
    get isObject(): boolean;
    /** Return a PL variable from this object. */
    toPLVariable: (options?: GenOptions) => PLVariableInfo;
}
/** Create a PL variable from a name, type and options. */
export declare function createPLVariable(name: string, type: GraphQLType, options?: GenOptions): PLVariableInfo;
/** Options for generating a programming language variable. */
export interface GenOptions {
    mkListZero: (tname: string) => string;
    mkOptZero: (tname: string) => string;
    /** Object types have a path prefix given by this value, if present. */
    objectsHaveParentType?: string;
    /** Final scalar map to use if scalar type. */
    scalars?: NormalizedScalarsMap;
    enums?: ParsedEnumValuesMap;
}
/** Default gen options. */
export declare const defaultGenOptions: GenOptions;
export declare const ArrayGenOptions: Partial<GenOptions>;
export declare const NullGenOptions: Partial<GenOptions>;
export declare const UndefinedGenOptions: Partial<GenOptions>;
