"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const visitor_plugin_common_1 = require("@graphql-codegen/visitor-plugin-common");
const graphql_1 = require("graphql");
const types_1 = require("./types");
const logger_1 = require("./logger");
/** Resolve AST FieldNodes to ResolvedFields. Throws an error if the
 * field is not a field on the parent type.
 */
function resolveFieldNodes(parent, fields) {
    const lookup = parent.getFields();
    return fields.map(f => {
        const type_name = f.name.value; // use name not alias!
        const field_def = lookup[type_name];
        if (!field_def)
            throw new Error(`Could not find ${type_name} field on ${parent.name}`);
        return ResolvedField.fromField(parent, field_def, f);
    });
}
exports.resolveFieldNodes = resolveFieldNodes;
/** Resolve a list of SelectionNodes (defined in operations) against
 * a parent and the overall schema.
 */
function resolveSelectionSet(context, parentType, set) {
    const logme = logger_1.log.extend("resolveSelectionSet");
    logme(`Resolving selection set on parent type ${parentType}. ${set.length} selection(s).`);
    const { fields, spreads } = visitor_plugin_common_1.separateSelectionSet(set);
    logme(`Inspecting fields: ${fields.length}, spreads: ${spreads.length}`);
    const resolved = resolveFieldNodes(parentType, fields);
    // convert spreads to fields which must be fields on this parent type
    const spread_fields = spreads.flatMap(s => resolveFragmentSpreadToFields(context, parentType, s));
    logme(`Generated ${resolved.length} fields from fields: (${resolved.map(f => f.name)})`);
    logme(`Generated ${spread_fields.length} fields from spreads: (${spread_fields.map(f => f.name)}).`);
    return new ResolvedSelectionSet(context.schema, parentType, [...resolved, ...spread_fields]);
}
exports.resolveSelectionSet = resolveSelectionSet;
/** Use fragment name on FragmentSpreadNode to create a list of resolved fields.
 * FragmentSpreadNodes are found in operation definitions and represent the use
 * of a spread vs the definition of a spread.
 *
 * @see resolveFragmentNameToFields
 */
function resolveFragmentSpreadToFields(context, parentType, spread) {
    const logme = logger_1.log.extend("resolveFragmentSpreadToFields");
    const spread_name = spread.name.value;
    const fields = resolveFragmentNameToFields(context, spread_name);
    logme(`Resolved spread ${spread_name} to ${fields.length} fields.`);
    if (fields[0].name !== parentType.name)
        throw new Error(`Resolved fragment named ${spread_name} appears to have a different type from ${parentType.name}`);
    return fields[1].map(f => { var _a, _b; return ResolvedField.from(parentType, f[1], aliasOrName(f[0]), (_b = (_a = f[0].selectionSet) === null || _a === void 0 ? void 0 : _a.selections) !== null && _b !== void 0 ? _b : []); });
}
exports.resolveFragmentSpreadToFields = resolveFragmentSpreadToFields;
/** Give a FieldNode, defined in selection set, return its alias or its name. */
function aliasOrName(f) {
    var _a, _b;
    return (_b = (_a = f.alias) === null || _a === void 0 ? void 0 : _a.value) !== null && _b !== void 0 ? _b : f.name.value;
}
exports.aliasOrName = aliasOrName;
/** Given the name of a fragment and the overall schema, expand
 * the fragment to a set of fields via the FragmentDefinition found in the context.
 * If any of the toplevel fragment
 * elements are fragments, expand those. Also return the corresponding FieldNode
 * which have additional information about aliases and directives.
 */
function resolveFragmentNameToFields(context, name) {
    const logme = logger_1.log.extend("resolveFragmentNameToFields");
    // find fragment definition
    const spread = context.fragmentByName(name);
    if (!spread)
        return null;
    const ofTypeName = spread.typeCondition.name.value;
    const ofType = context.schema.getType(ofTypeName);
    logme("ofType", ofType);
    logme("debug ofType");
    if (logger_1.log.enabled)
        types_1.debug_type(ofType);
    const get_field = (fname) => graphql_1.isObjectType(ofType) && ofType.getFields() ? ofType.getFields()[fname] : null;
    // process each field's AST
    const fields = spread.selectionSet.selections
        .map(sel => {
        switch (sel.kind) {
            case graphql_1.Kind.FIELD:
                // FieldNode
                const f = get_field(sel.name.value);
                return [[sel, f]];
                break;
            case graphql_1.Kind.FRAGMENT_SPREAD:
                // fragment spread within the top level of this fragment spread
                // means it must be spread on the same type as the current lever's spread
                const r = resolveFragmentNameToFields(context, sel.name.value);
                if (!r)
                    return null;
                // check that parents match...
                const check_type_name = r[0].name;
                if (check_type_name !== ofTypeName)
                    throw new Error(`In fragment ${name}(${ofTypeName}) top-level fragment expansion had non-matching type: ${check_type_name}.`);
                return r[1];
                break;
            case graphql_1.Kind.INLINE_FRAGMENT:
                throw new Error("resolveFragementSpreadAST: Kind.INLINE_FRAGMENT handling is unimplemented.");
                return null;
                break;
        }
    })
        .filter(a => a)
        .flat();
    return [ofType, fields];
}
exports.resolveFragmentNameToFields = resolveFragmentNameToFields;
/** Given a resolved field, resolve its selections. */
function resolveFieldSelectionSet(context, field) {
    const selectionSet = field.selectionSet;
    // field could be named/list/not-null, recurse to get core type
    const core_type = visitor_plugin_common_1.getBaseTypeNode(field.field.astNode.type);
    const field_type = context.schema.getType(core_type.name.value);
    if (!graphql_1.isObjectType(field_type))
        throw new Error(`Resolved field '${field_type}.${field.name}' does not have selects to resolve.`);
    return resolveSelectionSet(context, field_type, selectionSet);
}
exports.resolveFieldSelectionSet = resolveFieldSelectionSet;
/** Resolve selection sets to a hierarchical, intermediate data structure
 * useful for different target languages and generate scala code. This
 * function is typically used at the highest level of definition
 * for an operation and the parentSchemaType is typically Query, Mutation
 * or Subscription. It is a wrapper over `resolveSelectionSet`.
 */
function selectionSetToObject(context, parentSchemaType, selectionSet) {
    return graphql_1.isObjectType(parentSchemaType)
        ? resolveSelectionSet(context, parentSchemaType, selectionSet.selections)
        : null;
}
exports.selectionSetToObject = selectionSetToObject;
/** For all resolution, basic data that may be needed. */
class ResolveContext {
    constructor(schema, fragments, scalars) {
        this.schema = schema;
        this.fragments = fragments;
        this.scalars = scalars;
        /** Find a fragment by name. */
        this.fragmentByName = (name) => this.fragments.find(f => f.name.value === name);
    }
}
exports.ResolveContext = ResolveContext;
/** Intermediate representation of resolved fields for a given
 * type. Sub-objects can be resolved by resolving the "complex"
 * fields.
 */
class ResolvedSelectionSet {
    constructor(schema, ofType, fields) {
        this.schema = schema;
        this.ofType = ofType;
        this.fields = fields;
        /** This should resolve all objects and their resolved selection sets. */
        this.resolveComplex = (context) => {
            logger_1.log("Resolving complex fields", this.fields.map(f => `${f.name} (${f.field.type})`).join(", "));
            return this.complex.map(f => [f, resolveFieldSelectionSet(context, f)]);
        };
        if (!ofType)
            throw new Error(`Cannot create ResolvedSelectionSet with null ofType`);
    }
    get leaves() {
        return this.fields.filter(f => f.isScalar);
    }
    get complex() {
        return this.fields.filter(f => f.isObject);
    }
    /** Partition the resolved fields into leaves and non-leaves. */
    get partition() {
        return [this.leaves, this.complex];
    }
}
exports.ResolvedSelectionSet = ResolvedSelectionSet;
/** Intermediate represetation of a field selection from an object for use
 * in an operation such as a query.
 */
class ResolvedField {
    constructor(
    /** Name of field may be influenced by alias and not what "field" says. */
    name, 
    /** Field type. */
    field, 
    /** Parent type. */
    ofType, 
    /** Subfields, if any in AST format. They are not resolved. */
    selectionSet) {
        this.name = name;
        this.field = field;
        this.ofType = ofType;
        this.selectionSet = selectionSet;
        /** Return a PL variable from this object. */
        this.toPLVariable = (options) => {
            return createPLVariable(this.name, this.field.type, options);
        };
        if (logger_1.log.enabled) {
            logger_1.log(`Created ResolvedField ${name}`);
            types_1.debug_type(field.type);
        }
    }
    get isLeaf() {
        return graphql_1.isLeafType(graphql_1.getNamedType(this.field.type));
    }
    get isScalar() {
        return graphql_1.isScalarType(graphql_1.getNamedType(this.field.type));
    }
    get isObject() {
        return graphql_1.isObjectType(graphql_1.getNamedType(this.field.type));
    }
}
exports.ResolvedField = ResolvedField;
/** Same as constructor. */
ResolvedField.from = (ofType, field, name, selectionSet) => {
    return new ResolvedField(name, field, ofType, selectionSet);
};
/** FieldNodes are defined in operations. */
ResolvedField.fromField = (ofType, field, astNode) => {
    var _a, _b, _c, _d;
    return new ResolvedField((_b = (_a = astNode.alias) === null || _a === void 0 ? void 0 : _a.value) !== null && _b !== void 0 ? _b : field.name, field, ofType, (_d = (_c = astNode.selectionSet) === null || _c === void 0 ? void 0 : _c.selections) !== null && _d !== void 0 ? _d : []);
};
/** Create a PL variable from a name, type and options. */
function createPLVariable(name, type, options) {
    const opts = Object.assign(Object.assign({}, exports.defaultGenOptions), options);
    //debug_type(type);
    const named_type = graphql_1.getNamedType(type);
    const bottom_name = named_type.name;
    let tname = graphql_1.isScalarType(named_type) && opts.scalars ? options.scalars[bottom_name] : bottom_name;
    tname = `${graphql_1.isObjectType(named_type) && opts.objectsHaveParentType ? `${opts.objectsHaveParentType}.` : ""}${tname}`;
    let defaultValue = undefined;
    if (graphql_1.isListType(type))
        defaultValue = opts.mkListZero(tname);
    else if (!graphql_1.isNonNullType(type))
        defaultValue = opts.mkOptZero(tname);
    return {
        name: name,
        type: tname,
        wrapper: types_1.mk_type_wrapper_thunk(type),
        defaultValue
    };
}
exports.createPLVariable = createPLVariable;
/** Default gen options. */
exports.defaultGenOptions = {
    mkListZero: (tname) => undefined,
    mkOptZero: (tname) => undefined
};
exports.ArrayGenOptions = {
    mkListZero: (tname) => "js.Array()"
};
exports.NullGenOptions = {
    mkOptZero: (tname) => "null"
};
exports.UndefinedGenOptions = {
    mkOptZero: (name) => "js.undefined"
};
//# sourceMappingURL=SelectionSetToObject.js.map