import { OperationDefinitionNode, FragmentDefinitionNode } from "graphql";
import * as graphql from "graphql";
import OperationVariablesToObject from "./OperationVariablesToObject";
import { ResolveContext, ResolvedSelectionSet } from "./SelectionSetToObject";
import { Config } from "./config";
import { LoadedFragment, ConvertOptions, ConvertFn } from "@graphql-codegen/visitor-plugin-common";
/** Generate generic scala.js graphql data and data structures.
 * Each operation becomes one scala object.
 *
 * Unlike some of the other @graphql-codegen visitors we need to integrate
 * a "client" visitor for running over types as well as a "documents"
 * visitor to match up schema types so we can output both gql strings
 * for each operation but also the data structures which we design to be unique to each
 * operation. All fragments are grouped together
 * into a `Fragments` object that is referenced directly. We
 * make our own visitor object because class inherintance seemed messy and there
 * was not alot of code resue given the assumptions made in the provided
 * visitors.
 */
export declare class ScalaJSVisitor {
    _config: Config;
    protected _operationNameMappings: Record<string, string>;
    protected _variablesTransformer: OperationVariablesToObject;
    protected _additionalImports: Array<string>;
    protected _collectedOperations: OperationDefinitionNode[];
    constructor(_config: Config);
    get config(): Config;
    /** Output raw document. Must have gqlImport propName defined to produce gql function call. */
    protected _buildDocumentNodes: (node: any) => string;
    protected _buildOperationResultTypes: (node: OperationDefinitionNode, operationRootType: graphql.GraphQLObjectType<any, any, {
        [key: string]: any;
    }>) => string;
    protected _buildOperationVariables: (node: OperationDefinitionNode) => string;
    protected _buildScalaContainer: (node: OperationDefinitionNode, oname: string, operationRootType: graphql.GraphQLObjectType<any, any, {
        [key: string]: any;
    }>) => string;
    /** The superclass method generates a document string that is not relevant to us. */
    OperationDefinition: (node: OperationDefinitionNode) => string;
    /** Comment with all operation names listed so you know what's here. */
    get op_comment(): string;
}
/** scala: Generate scala imports statements from a list of import strings, either direct scala
 * packages or in the form moduleName#propName (only moduleName is retained).
 */
export declare function genImports(list: Array<string>): string[];
/** Generate DocumentNode content with embedded fragment "string" includes.
 * The node parameter is searched downward for fragments and only those fragment
 * names are embedded into the document string representation. The generation
 * method assumes that embedded fragment names are prefixed by `fragmentObjectName`
 * and all fragments are found in subnodes of `node.`.
 *
 * @param node Content to convert to a string.
 * @param fragmentObjectName Prefix to fragment names embeded in GQL documents.
 * @param convertName Convert fragment names found in node to match what might be generated
 * elsewhere in the code base.
 */
export declare function makeGQLForScala(node: FragmentDefinitionNode | OperationDefinitionNode, fragmentObjectName: string, convertName?: ConvertFn<ConvertOptions>): string;
/** Extract fragments from document, use thunk to create language specific
 * embeddings referencing those fragments and generate the string version of the document.
 */
export declare function makeGQLWithThunk(node: FragmentDefinitionNode | OperationDefinitionNode, thunk: (fname: string) => string): string;
/** Make GQL string from a node embedding the fragment references. The
 * fragment references should be output language specific.
 */
export declare function makeGQLWithFragments(node: FragmentDefinitionNode | OperationDefinitionNode, fragmentReferences: Array<string>): string;
/** scala: Scan LoadedFragments for all fragments and generate a scala object
 * containing their declarations as a DocumentNode via "gql".
 */
export declare function getFragmentBlock(allFragments: Array<LoadedFragment>, objectName: string, makeVariableName: (node: FragmentDefinitionNode) => string, makeGQL: (node: FragmentDefinitionNode) => string): string;
/** If operation is anonymous, generate a synthensized operation name otherwise
 * return the raw operation name.
 * Return tuple [name was synthensized, name]. Side effect of incrementing
 * hiddent state counter.
 */
export declare function generateOperationName(node: OperationDefinitionNode): [boolean, string];
export interface RecurseContext {
    path: Array<string>;
}
/** Render traits recursively, output hierarchical/dependent types. */
export declare function renderRecursiveData(selects: ResolvedSelectionSet, ctx: ResolveContext, state: RecurseContext): any;
