"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const graphql_1 = require("graphql");
const graphql = __importStar(require("graphql"));
const pascal_case_1 = require("pascal-case");
const OperationVariablesToObject_1 = __importDefault(require("./OperationVariablesToObject"));
const types_1 = require("./types");
const SelectionSetToObject_1 = require("./SelectionSetToObject");
const config_1 = require("./config");
const fragments_1 = require("./fragments");
const logger_1 = require("./logger");
const trait_1 = require("./trait");
/** Generate generic scala.js graphql data and data structures.
 * Each operation becomes one scala object.
 *
 * Unlike some of the other @graphql-codegen visitors we need to integrate
 * a "client" visitor for running over types as well as a "documents"
 * visitor to match up schema types so we can output both gql strings
 * for each operation but also the data structures which we design to be unique to each
 * operation. All fragments are grouped together
 * into a `Fragments` object that is referenced directly. We
 * make our own visitor object because class inherintance seemed messy and there
 * was not alot of code resue given the assumptions made in the provided
 * visitors.
 */
class ScalaJSVisitor {
    constructor(_config) {
        this._config = _config;
        this._operationNameMappings = {};
        this._collectedOperations = [];
        /** Output raw document. Must have gqlImport propName defined to produce gql function call. */
        this._buildDocumentNodes = (node) => {
            const raw = makeGQLForScala(node, this.config.fragmentObjectName, this.config.convertName);
            const rawOp = `val operationString = """${raw}"""`;
            let docOp = "";
            if (this.config.gqlImport) {
                const parsed = config_1.parseImport(this.config.gqlImport);
                if (parsed.propName) {
                    docOp = `val operation = ${parsed.propName}(operationString)`;
                }
            }
            return [rawOp, docOp].filter(a => a).join("\n");
        };
        this._buildOperationResultTypes = (node, operationRootType) => {
            const logme = logger_1.log.extend("_buildOperationResultType");
            const ctx = new SelectionSetToObject_1.ResolveContext(this._config.schema, config_1.fragmentDefinitions(this._config.fragments), this._config.scalars);
            const resolved = SelectionSetToObject_1.selectionSetToObject(ctx, operationRootType, node.selectionSet);
            logme("Resolved selection set: type %O, fields: %O (simple: %d, complex: %d)", resolved.ofType, resolved.fields, resolved.leaves.length, resolved.complex.length);
            const genopts = Object.assign(Object.assign(Object.assign({}, SelectionSetToObject_1.NullGenOptions), SelectionSetToObject_1.ArrayGenOptions), { scalars: this.config.scalars, objectsHaveParentType: "Data" });
            const simpleFields = resolved ? resolved.leaves.map(f => f.toPLVariable(genopts)) : [];
            const subObjects = resolved ? resolved.resolveComplex(ctx) : [];
            const subObjectFields = subObjects.map(p => p[0].toPLVariable(Object.assign({}, genopts)));
            const nested = subObjects
                .map(so => renderRecursiveData(so[1], ctx, {
                path: ["Data"]
            }))
                .join("\n");
            logme(`${simpleFields.length} simple fields, ${subObjects.length} sub-object fields.`);
            logme(`Toplevel Data object direct fields: ${simpleFields.map(f => f.name)}`);
            logme(`   Subobject fields: ${subObjectFields.map(f => f.name)}`);
            return trait_1.generateScalaJSTrait("Data", [...simpleFields, ...subObjectFields], {
                native: true,
                nested
            });
        };
        this._buildOperationVariables = (node) => {
            if (node.variableDefinitions.length > 0) {
                const operationVariables = this._variablesTransformer.transform(node.variableDefinitions);
                return trait_1.generateScalaJSTrait("Variables", operationVariables);
            }
            return "";
        };
        this._buildScalaContainer = (node, oname, operationRootType) => {
            try {
                return trait_1.generateScalaObject(oname, [
                    this._buildDocumentNodes(node),
                    this._buildOperationVariables(node),
                    this._buildOperationResultTypes(node, operationRootType)
                ].join("\n"));
            }
            catch (e) {
                logger_1.log("Error generating scala container", e);
                return e.message;
            }
        };
        /** The superclass method generates a document string that is not relevant to us. */
        this.OperationDefinition = (node) => {
            // for dev, only process a few entries...
            //if (this._collectedOperations.length > 2) return null;
            this._collectedOperations.push(node);
            const [sythesized_op_name, op_name] = generateOperationName(node);
            // schema's listof all possible operations
            const operationRootType = types_1.getRootType(node.operation, this.config.schema);
            if (!operationRootType)
                throw new Error(`Unable to find root schema for operation type "${node.operation}"`);
            const operationRootSuffix = pascal_case_1.pascalCase(operationRootType.name);
            // outer object container name with operation type suffix if it's not there
            const objectNameSuffix = op_name
                .toLowerCase()
                .endsWith(operationRootSuffix.toLowerCase())
                ? ""
                : operationRootSuffix;
            const objectName = this.config.convertName(op_name, {
                suffix: objectNameSuffix
            });
            const theObject = this._buildScalaContainer(node, objectName, operationRootType);
            this._operationNameMappings[op_name] = objectName;
            return [theObject].filter(a => a).join("\n");
        };
        this._additionalImports = ["scala.scalajs.js"];
        const enumsNames = Object.keys(_config.schema.getTypeMap()).filter(tn => graphql_1.isEnumType(_config.schema.getType(tn)));
        this._variablesTransformer = new OperationVariablesToObject_1.default(_config.scalars, _config.convertName.bind(this), enumsNames, false, this.config.enumValues);
    }
    get config() {
        return this._config;
    }
    /** Comment with all operation names listed so you know what's here. */
    get op_comment() {
        return ["// ops mappings:"]
            .concat(Object.entries(this._operationNameMappings).map(n => `// ${n[0]} => ${n[1]}`))
            .join("\n");
    }
}
exports.ScalaJSVisitor = ScalaJSVisitor;
/** scala: Generate scala imports statements from a list of import strings, either direct scala
 * packages or in the form moduleName#propName (only moduleName is retained).
 */
function genImports(list) {
    return list
        .map(config_1.parseImport)
        .map(s => s.moduleName)
        .map(i => `import ${i}`);
}
exports.genImports = genImports;
/** Generate DocumentNode content with embedded fragment "string" includes.
 * The node parameter is searched downward for fragments and only those fragment
 * names are embedded into the document string representation. The generation
 * method assumes that embedded fragment names are prefixed by `fragmentObjectName`
 * and all fragments are found in subnodes of `node.`.
 *
 * @param node Content to convert to a string.
 * @param fragmentObjectName Prefix to fragment names embeded in GQL documents.
 * @param convertName Convert fragment names found in node to match what might be generated
 * elsewhere in the code base.
 */
function makeGQLForScala(node, fragmentObjectName, convertName) {
    const fragments = fragments_1.extractFragments(node);
    const thunk = (fname) => genScalaFragmentReferencesForDocument(fragmentObjectName, convertName ? convertName(fname) : fname);
    return makeGQLWithThunk(node, thunk);
}
exports.makeGQLForScala = makeGQLForScala;
/** Extract fragments from document, use thunk to create language specific
 * embeddings referencing those fragments and generate the string version of the document.
 */
function makeGQLWithThunk(node, thunk) {
    const fragment_names = fragments_1.extractFragments(node);
    const deduped = [...new Set(fragment_names)];
    const refs = deduped.map(fn => thunk(fn));
    return makeGQLWithFragments(node, refs);
}
exports.makeGQLWithThunk = makeGQLWithThunk;
/** Make GQL string from a node embedding the fragment references. The
 * fragment references should be output language specific.
 */
function makeGQLWithFragments(node, fragmentReferences) {
    const fragments = fragments_1.extractFragments(node);
    const doc = `${graphql
        .print(node)
        .split("\\")
        .join("\\\\")}
            ${fragmentReferences.join("\n")}`;
    return doc;
}
exports.makeGQLWithFragments = makeGQLWithFragments;
/** scala: Scan LoadedFragments for all fragments and generate a scala object
 * containing their declarations as a DocumentNode via "gql".
 */
function getFragmentBlock(allFragments, objectName, makeVariableName, makeGQL) {
    const fragments = fragments_1.getFragments(allFragments).map(doc => `val ${makeVariableName(doc)} = """"${makeGQL(doc)}"""`);
    return trait_1.generateScalaObject(objectName, fragments.join("\n"));
}
exports.getFragmentBlock = getFragmentBlock;
/** scala: Specific to scala string templates, given fragment string names,
 * generate content for
 * placing fragment names into a gql document string.
 */
function genScalaFragmentReferencesForDocument(prefix, fname) {
    return "${" + prefix + "." + fname + "}";
}
/** Obtain a list of operations in the document. (non-mutating) */
function findOperations(doc) {
    const nodes = [];
    const result = graphql_1.visit(doc, {
        leave: {
            OperationDefinition(node) {
                nodes.push(node);
            }
        }
    });
    return nodes;
}
let _unnamedCounter = 1;
/** If operation is anonymous, generate a synthensized operation name otherwise
 * return the raw operation name.
 * Return tuple [name was synthensized, name]. Side effect of incrementing
 * hiddent state counter.
 */
function generateOperationName(node) {
    const name = node.name && node.name.value;
    if (name) {
        return [false, name];
    }
    return [true, `Unnamed_${_unnamedCounter++}_`];
}
exports.generateOperationName = generateOperationName;
/** Render traits recursively, output hierarchical/dependent types. */
function renderRecursiveData(selects, ctx, state) {
    if (!selects)
        throw new Error("selects is null");
    const logme = logger_1.log.extend("_renderRecursiveData");
    const name = selects.ofType.name;
    const newstate = Object.assign(Object.assign({}, state), { path: state.path.concat([name]) });
    const path = state.path.join(".");
    const fqn = `${path}.${name}`;
    const genopts = Object.assign(Object.assign(Object.assign({}, SelectionSetToObject_1.NullGenOptions), SelectionSetToObject_1.ArrayGenOptions), { scalars: ctx.scalars, objectsHaveParentType: newstate.path.join(".") });
    const directs = selects.leaves.map(f => f.toPLVariable(genopts));
    logme(`Generating trait ${path}.${name}`);
    const subObjects = selects.resolveComplex(ctx);
    const subObjectFields = subObjects.map(s => s[0].toPLVariable(genopts));
    const nested = subObjects.map(p => renderRecursiveData(p[1], ctx, newstate)).join("\n");
    return trait_1.generateScalaJSTrait(name, [...directs, ...subObjectFields], {
        native: true,
        nested,
        fqn
    });
}
exports.renderRecursiveData = renderRecursiveData;
//# sourceMappingURL=operations.js.map