"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const graphql_1 = require("graphql");
const logger_1 = require("./logger");
/** Given an OperationTypeNode , return the root (query,mutation,subscription) in the schema. */
function getRootType(operation, schema) {
    switch (operation) {
        case "query":
            return schema.getQueryType();
        case "mutation":
            return schema.getMutationType();
        case "subscription":
            return schema.getSubscriptionType();
    }
}
exports.getRootType = getRootType;
/*** Debug print a type using recursion to unwrap if needed. */
function debug_type(t, indent = 0, log = console.log) {
    const i = new Array(indent + 1).fill("  ").join("");
    const start = new Array(indent).fill("  ").join("");
    log(`${start}debug of type: '${t}'`);
    log(`${i}isInput`, graphql_1.isInputType(t));
    log(`${i}isOutput`, graphql_1.isOutputType(t));
    log(`${i}isNullable`, graphql_1.isNullableType(t));
    log(`${i}isNonNull`, graphql_1.isNonNullType(t));
    log(`${i}isList`, graphql_1.isListType(t));
    log(`${i}isWrapping`, graphql_1.isWrappingType(t));
    log(`${i}isLeaf`, graphql_1.isLeafType(t));
    log(`${i}isObject`, graphql_1.isObjectType(t));
    log(`${i}isScalar`, graphql_1.isScalarType(t));
    log(`${i}isComposite`, graphql_1.isCompositeType(t));
    log(`${i}isAbstract`, graphql_1.isAbstractType(t));
    log(`${i}isEnum`, graphql_1.isEnumType(t));
    log(`${i}getNamedType`, graphql_1.getNamedType(t));
    log(`${i}getNullableType`, graphql_1.getNullableType(t));
    if (graphql_1.isWrappingType(t))
        debug_type(t.ofType, indent + 1);
}
exports.debug_type = debug_type;
exports.defaultWrapperOptions = {
    mkList: (arg) => `js.Array[${arg}]`,
    mkOpt: (arg) => `${arg}|Null`,
    mkOptZero: () => "null",
    mkListZero: () => "js.Array()"
};
/** Build a rendering thunk for the type. Takes
 * into account null and list types only i.e. wrapping types.
 */
function mk_type_wrapper_thunk(t, options) {
    if (!t)
        throw Error("Type was null in type wrapper generation");
    options = Object.assign(Object.assign({}, exports.defaultWrapperOptions), options);
    const original_type = t;
    // if true then lt = false and nnt = false
    let w = graphql_1.isWrappingType(t);
    // these are hierarchical, its one or the other
    let lt = graphql_1.isListType(t);
    let nnt = graphql_1.isNonNullType(t);
    let ofType = graphql_1.isWrappingType(t) ? t.ofType : null;
    let ofTypeW = ofType ? graphql_1.isWrappingType(ofType) : false;
    // should never happen
    if (graphql_1.isWrappingType(t) && !ofType)
        throw new Error(`Wrapping type ${t} but no ofType defined in type wrapper generation.`);
    const ops = [];
    if (!w)
        ops.push("N");
    while (w) {
        let r = "";
        if (!nnt && !lt)
            r = "N";
        else if (lt)
            r = "L";
        else
            r = "-";
        logger_1.log("determining type wrapper op", t, "wt:", w, ", lt:", lt, ", nnt:", nnt, ", ofType:", ofType, ", ofType wt:", ofTypeW, ", result => ", r);
        ops.push(r);
        t = ofType;
        w = t ? graphql_1.isWrappingType(t) : false;
        if (t) {
            ofType = graphql_1.isWrappingType(t) ? t.ofType : null;
            lt = graphql_1.isListType(t);
            nnt = graphql_1.isNonNullType(t);
        }
        else {
            ofType = null;
            lt = false;
            nnt = false;
        }
    }
    return (targ) => {
        const r = ops.reduce((p, c) => {
            switch (c) {
                case "N":
                    return options.mkOpt(p);
                case "L":
                    return options.mkList(p);
                default:
                    return p;
            }
        }, targ);
        logger_1.log(`appyling type wrapper ops: original type '${original_type}'`, "arg", targ, "ops", ops, " => ", r);
        return r;
    };
}
exports.mk_type_wrapper_thunk = mk_type_wrapper_thunk;
//# sourceMappingURL=types.js.map