import {
  convertFactory,
  NamingConvention,
  ConvertOptions,
  ConvertFn,
  buildScalars,
  ScalarsMap,
  NormalizedScalarsMap,
  EnumValuesMap,
  ParsedEnumValuesMap,
  LoadedFragment
} from "@graphql-codegen/visitor-plugin-common";
import { GraphQLSchema } from "graphql";
import { DEFAULT_SCALARS } from "./scalars";
import { parseEnumValues } from "./enum-values";

/** Raw configuration filled in by the @graphql-codgen layer. */
export interface RawConfig {
  /** For declaring documents in scala objects. If missing, documents are not added. */
  gqlImport?: string;
  /** Applies to typeNames and enumValues. Example, lower-case#lowerCase or upper-case#upperCase. Default is pascal-case#pascalCase. */
  namingConvention?: NamingConvention;
  fragmentObjectName?: string;
  scalars?: ScalarsMap;
  /** Does not add __typename to the generated type, unless it was specified. Default is false. */
  skipTypename?: boolean;
  /** Override enum values in the config. */
  enumValues?: EnumValuesMap;
  /** Externally loaded fragments. */
  externalFragments?: LoadedFragment[];
}

/** Final processed configuration taking into account defaults. */
export interface Config {
  schema: GraphQLSchema;
  gqlImport: string | null;
  convertName: ConvertFn<ConvertOptions>;
  fragmentObjectName: string;
  scalars: NormalizedScalarsMap;
  skipTypeName: boolean;
  enumValues: ParsedEnumValuesMap;
  /** All fragments, side-loaded as well as originally in the op docs. */
  fragments: LoadedFragment[];
}

/** Given a list of LoadedFragment objects, return the raw FragmentDefinitionNode AST
 * objects.
 */
export function fragmentDefinitions(lf: ReadonlyArray<LoadedFragment>) {
  return lf.map(lf => lf.node);
}

/** Make a Config from a RawConfig. */
export function makeConfig(schema: GraphQLSchema, raw: RawConfig): Config {
  const parsedScalarsMap = buildScalars(schema, raw.scalars, DEFAULT_SCALARS);
  const scalars = {};
  Object.keys(parsedScalarsMap).forEach(key => {
    scalars[key] = parsedScalarsMap[key].type;
  });
  return {
    schema,
    gqlImport: raw.gqlImport || null,
    convertName: convertFactory(raw),
    fragmentObjectName: raw.fragmentObjectName || "Fragment",
    scalars,
    skipTypeName: raw.skipTypename || false,
    enumValues: parseEnumValues(schema, raw.enumValues),
    fragments: raw.externalFragments || []
  };
}

/** Parse an import config element in the form moduleName#propName */
export function parseImport(s: string) {
  const [mname, pname] = s.split("#");
  return {
    moduleName: mname,
    propName: pname
  };
}
