import { LoadedFragment } from "@graphql-codegen/visitor-plugin-common";
import { DepGraph } from "dependency-graph";
import {
  OperationDefinitionNode,
  FragmentDefinitionNode,
  FragmentSpreadNode,
  visit,
  print
} from "graphql";

/** Find all FragmentDefinitionNode children. Dedup circles using a dependency graph. */
export function getFragments(
  fragments: Array<LoadedFragment>
): FragmentDefinitionNode[] {
  const graph = fragmentsGraph(fragments);
  const orderedDeps = graph.overallOrder();
  const localFragments = orderedDeps.filter(
    name => !graph.getNodeData(name).isExternal
  );
  return localFragments.map(name => graph.getNodeData(name).node);
}

/** Generate graph of names=>document content mappings. */
export function fragmentsGraph(
  fragments: LoadedFragment[]
): DepGraph<LoadedFragment> {
  const graph = new DepGraph<LoadedFragment>({ circular: true });

  for (const fragment of fragments) {
    if (graph.hasNode(fragment.name)) {
      const cachedAsString = print(graph.getNodeData(fragment.name).node);
      const asString = print(fragment.node);

      if (cachedAsString !== asString) {
        throw new Error(`Duplicated fragment called '${fragment.name}'!`);
      }
    }

    graph.addNode(fragment.name, fragment);
  }

  fragments.forEach(fragment => {
    const depends = extractFragments(fragment.node);

    if (depends && depends.length > 0) {
      depends.forEach(name => {
        graph.addDependency(fragment.name, name);
      });
    }
  });

  return graph;
}

/** Extract out all FragmentSpread names in children AST nodes. */
export function extractFragments(
  document: FragmentDefinitionNode | OperationDefinitionNode
): string[] {
  if (!document) {
    return [];
  }

  const names = [];

  visit(document, {
    enter: {
      FragmentSpread: (node: FragmentSpreadNode) => {
        names.push(node.name.value);
      }
    }
  });

  return names;
}
