// https://graphql-code-generator.com/docs/custom-codegen/write-your-plugin
import "source-map-support/register";
import { PluginFunction, PluginValidateFn, Types } from "@graphql-codegen/plugin-helpers";
import { LoadedFragment } from "@graphql-codegen/visitor-plugin-common";
import { GraphQLSchema, visit, concatAST, FragmentDefinitionNode, Kind } from "graphql";
import {
  ScalaJSVisitor,
  getFragmentBlock as genFragmentBlock,
  makeGQLForScala,
  genImports
} from "./operations";
import { RawConfig, makeConfig } from "./config";
import { genEnums } from "./enums";

export const plugin: PluginFunction<RawConfig> = (
  schema: GraphQLSchema,
  documents: Types.DocumentFile[],
  config: RawConfig
) => {
  const allAst = concatAST(documents.map(v => v.document));

  // Gather all fragments into one place since finding them on the
  // fly in the schema is painful.
  const allFragments: LoadedFragment[] = [
    ...(allAst.definitions.filter(
      d => d.kind === Kind.FRAGMENT_DEFINITION
    ) as FragmentDefinitionNode[]).map(fragmentDef => ({
      node: fragmentDef,
      name: fragmentDef.name.value,
      onType: fragmentDef.typeCondition.name.value,
      isExternal: false
    })),
    ...(config.externalFragments || [])
  ];

  // console.log("all types");
  // Object.keys(schema.getTypeMap())
  //   .sort()
  //   .forEach(k => {
  //     const t = schema.getType(k);
  //     console.log("type: ", k, ": ", t);
  //   });

  try {
    const final_config = makeConfig(schema, {
      ...config,
      externalFragments: allFragments
    });
    //const ops = extractOperations(allAst).map(
    const visitor = new ScalaJSVisitor(final_config);
    const visitorResult = visit(allAst, { leave: visitor as any });

    return {
      prepend: [
        ...genImports(["scala.scalajs.js", "js.|", final_config.gqlImport]),
        visitor.op_comment
      ],
      content: [
        // `
        // object __Conversions {
        //   implicit class RichOrNull[T](val v: T |Null) {
        //     def forceGet = if(v == null) null else v.asInstanceOf[T]
        //   }
        // }
        // import __Conversions._
        // `,
        genEnums(schema),
        genFragmentBlock(
          final_config.fragments,
          final_config.fragmentObjectName,
          final_config.convertName,
          node => makeGQLForScala(node, final_config.fragmentObjectName, final_config.convertName)
        ),
        ...visitorResult.definitions.filter(t => typeof t === "string")
      ].join("\n")
    };
  } catch (e) {
    console.log("Error generating content", e);
    return { content: `Error generating content ${e.message}` };
  }
};

export const validate: PluginValidateFn<any> = async (
  schema: GraphQLSchema,
  documents: Types.DocumentFile[],
  config: any,
  outputFile: string
) => {
  if (!outputFile.endsWith(".scala")) {
    throw new Error(`Plugin "scalajs-client" requires output file extension to be ".scala"!`);
  }
};
