import {
  OperationTypeNode,
  GraphQLSchema,
  GraphQLType,
  isInputType,
  isOutputType,
  isNullableType,
  isNonNullType,
  isListType,
  isWrappingType,
  isLeafType,
  isObjectType,
  isScalarType,
  isCompositeType,
  isAbstractType,
  isEnumType,
  getNamedType,
  getNullableType
} from "graphql";
import { log } from "./logger";

/** Programming language variable broken out so it can be
 * used in multiple contexts.
 */
export interface PLVariableInfo {
  /** Varible name. */
  readonly name: string;
  /** Base type without wrapping for null/optionality. May have
   * list wrapper though so it may not be the element type itself.
   */
  readonly type: string;
  /** Call the wrapper to generate the full type signature. */
  readonly wrapper: (type: string) => string;
  /** Default value, raw string. scala may need converters here. */
  readonly defaultValue?: string;
}

/** Given an OperationTypeNode , return the root (query,mutation,subscription) in the schema. */
export function getRootType(operation: OperationTypeNode, schema: GraphQLSchema) {
  switch (operation) {
    case "query":
      return schema.getQueryType();
    case "mutation":
      return schema.getMutationType();
    case "subscription":
      return schema.getSubscriptionType();
  }
}

/*** Debug print a type using recursion to unwrap if needed. */
export function debug_type(t: GraphQLType, indent: number = 0, log = console.log) {
  const i = new Array(indent + 1).fill("  ").join("");
  const start = new Array(indent).fill("  ").join("");
  log(`${start}debug of type: '${t}'`);
  log(`${i}isInput`, isInputType(t));
  log(`${i}isOutput`, isOutputType(t));
  log(`${i}isNullable`, isNullableType(t));
  log(`${i}isNonNull`, isNonNullType(t));
  log(`${i}isList`, isListType(t));
  log(`${i}isWrapping`, isWrappingType(t));
  log(`${i}isLeaf`, isLeafType(t));
  log(`${i}isObject`, isObjectType(t));
  log(`${i}isScalar`, isScalarType(t));
  log(`${i}isComposite`, isCompositeType(t));
  log(`${i}isAbstract`, isAbstractType(t));
  log(`${i}isEnum`, isEnumType(t));
  log(`${i}getNamedType`, getNamedType(t));
  log(`${i}getNullableType`, getNullableType(t));
  if (isWrappingType(t)) debug_type(t.ofType, indent + 1);
}

export interface WrapperOptions {
  mkList: (arg: string) => string;
  mkOpt: (arg: string) => string;
  mkOptZero: () => string;
  mkListZero: () => string;
}

export const defaultWrapperOptions: WrapperOptions = {
  mkList: (arg: string) => `js.Array[${arg}]`,
  mkOpt: (arg: string) => `${arg}|Null`,
  mkOptZero: () => "null",
  mkListZero: () => "js.Array()"
};

/** Build a rendering thunk for the type. Takes
 * into account null and list types only i.e. wrapping types.
 */
export function mk_type_wrapper_thunk(
  t: GraphQLType,
  options?: Partial<WrapperOptions>
): (arg: string) => string {
  if (!t) throw Error("Type was null in type wrapper generation");
  options = { ...defaultWrapperOptions, ...options };
  const original_type = t;
  // if true then lt = false and nnt = false
  let w = isWrappingType(t);
  // these are hierarchical, its one or the other
  let lt = isListType(t);
  let nnt = isNonNullType(t);
  let ofType: GraphQLType | null = isWrappingType(t) ? t.ofType : null;
  let ofTypeW = ofType ? isWrappingType(ofType) : false;

  // should never happen
  if (isWrappingType(t) && !ofType)
    throw new Error(`Wrapping type ${t} but no ofType defined in type wrapper generation.`);

  const ops = [];

  if (!w) ops.push("N");
  while (w) {
    let r = "";
    if (!nnt && !lt) r = "N";
    else if (lt) r = "L";
    else r = "-";
    log(
      "determining type wrapper op",
      t,
      "wt:",
      w,
      ", lt:",
      lt,
      ", nnt:",
      nnt,
      ", ofType:",
      ofType,
      ", ofType wt:",
      ofTypeW,
      ", result => ",
      r
    );
    ops.push(r);
    t = ofType;
    w = t ? isWrappingType(t) : false;
    if (t) {
      ofType = isWrappingType(t) ? t.ofType : null;
      lt = isListType(t);
      nnt = isNonNullType(t);
    } else {
      ofType = null;
      lt = false;
      nnt = false;
    }
  }

  return (targ: string) => {
    const r = ops.reduce((p, c) => {
      switch (c) {
        case "N":
          return options.mkOpt(p);
        case "L":
          return options.mkList(p);
        default:
          return p;
      }
    }, targ);
    log(
      `appyling type wrapper ops: original type '${original_type}'`,
      "arg",
      targ,
      "ops",
      ops,
      " => ",
      r
    );
    return r;
  };
}
